"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const process_1 = require("@mesg/api/lib/process");
const decode_1 = __importDefault(require("./decode"));
const validate_1 = __importDefault(require("./validate"));
const schema = require('./schema/process.json');
const compileInstance = (def, opts) => __awaiter(void 0, void 0, void 0, function* () {
    return opts.instanceResolver
        ? yield opts.instanceResolver(def)
        : def.instanceHash;
});
const compileResult = (def, opts) => __awaiter(void 0, void 0, void 0, function* () {
    return ({
        type: "mesg.types.Process_Node_Result_",
        value: {
            result: {
                instanceHash: yield compileInstance(def, opts),
                taskKey: def.taskKey
            }
        }
    });
});
const compileEvent = (def, opts) => __awaiter(void 0, void 0, void 0, function* () {
    return ({
        type: "mesg.types.Process_Node_Event_",
        value: {
            event: {
                instanceHash: yield compileInstance(def, opts),
                eventKey: def.eventKey
            }
        }
    });
});
const compileTask = (def, opts) => __awaiter(void 0, void 0, void 0, function* () {
    return ({
        type: "mesg.types.Process_Node_Task_",
        value: {
            task: {
                instanceHash: yield compileInstance(def, opts),
                taskKey: def.taskKey
            }
        }
    });
});
const extractPathsFromKey = (key) => {
    return key.replace(/\[(\d)\]/g, val => `.${val}`).split('.');
};
const extractPathFromPaths = (paths) => {
    if (paths.length === 0)
        return null;
    const [currentPath, ...nextPaths] = paths;
    const indexRegex = /^\[(\d)\]$/;
    if (indexRegex.test(currentPath)) {
        const index = indexRegex.exec(currentPath)[1];
        const value = index && index !== '0'
            ? { index }
            : {};
        return {
            path: extractPathFromPaths(nextPaths),
            Selector: {
                type: "mesg.types.Process_Node_Reference_Path_Index",
                value
            }
        };
    }
    return {
        path: extractPathFromPaths(nextPaths),
        Selector: {
            type: 'mesg.types.Process_Node_Reference_Path_Key',
            value: {
                key: currentPath
            }
        }
    };
};
const compileMapOutput = (def, opts) => {
    if (def === null)
        return { Value: { type: "mesg.types.Process_Node_Map_Output_Null_", value: {} } };
    if (typeof def === 'number')
        return { Value: { type: "mesg.types.Process_Node_Map_Output_DoubleConst", value: { double_const: def } } };
    if (typeof def === 'boolean')
        return { Value: { type: 'mesg.types.Process_Node_Map_Output_BoolConst', value: { bool_const: def } } };
    if (typeof def === 'string')
        return { Value: { type: 'mesg.types.Process_Node_Map_Output_StringConst', value: { string_const: def } } };
    if (typeof def === 'object' && !!def.key)
        return {
            Value: {
                type: 'mesg.types.Process_Node_Reference',
                value: {
                    ref: {
                        path: extractPathFromPaths(extractPathsFromKey(def.key)),
                        nodeKey: def.stepKey || opts.defaultNodeKey
                    }
                }
            }
        };
    if (Array.isArray(def))
        return {
            Value: {
                type: 'mesg.types.Process_Node_Map_Output_List_',
                value: {
                    list: {
                        outputs: def.map(item => compileMapOutput(item, opts))
                    }
                }
            }
        };
    return {
        Value: {
            type: 'mesg.types.Process_Node_Map_Output_Map_',
            value: {
                map: Object.keys(def).map((key) => ({
                    Key: key,
                    Value: compileMapOutput(def[key], opts)
                })).sort((a, b) => a.Key.localeCompare(b.Key))
            }
        }
    };
};
const compileMap = (def, opts) => __awaiter(void 0, void 0, void 0, function* () {
    if (typeof def !== 'object' || Array.isArray(def))
        throw new Error('Map output should be a map');
    const outputs = compileMapOutput(def, opts);
    if (outputs.Value.type !== 'mesg.types.Process_Node_Map_Output_Map_')
        throw new Error('Map output should be a map');
    return {
        type: 'mesg.types.Process_Node_Map_',
        value: {
            map: outputs.Value.value.map
                .sort((a, b) => a.Key.localeCompare(b.Key))
        }
    };
});
const compileFilterValue = (def) => {
    if (def === null)
        return { type: "mesg.types.Value_NullValue", value: {} };
    if (typeof def === 'number')
        return { type: "mesg.types.Value_NumberValue", value: { number_value: def } };
    if (typeof def === 'boolean')
        return { type: 'mesg.types.Value_BoolValue', value: { bool_value: def } };
    if (typeof def === 'string')
        return { type: 'mesg.types.Value_StringValue', value: { string_value: def } };
    throw new Error('unsupported filter value');
};
const compileFilter = (def, opts) => __awaiter(void 0, void 0, void 0, function* () {
    const conditions = Array.isArray(def.conditions)
        ? def.conditions
        : Object.keys(def.conditions).sort((a, b) => a.localeCompare(b)).map(x => ({
            key: x,
            predicate: 'EQ',
            value: def.conditions[x]
        }));
    return {
        type: 'mesg.types.Process_Node_Filter_',
        value: {
            filter: {
                conditions: conditions
                    .map((condition) => ({
                    ref: {
                        path: extractPathFromPaths(extractPathsFromKey(condition.key)),
                        nodeKey: def.stepKey || opts.defaultNodeKey
                    },
                    predicate: process_1.Predicate[condition.predicate],
                    value: {
                        Kind: compileFilterValue(condition.value)
                    }
                }))
            }
        }
    };
});
const nodeCompiler = (type, def, key, opts) => __awaiter(void 0, void 0, void 0, function* () {
    const nodes = {
        result: compileResult,
        event: compileEvent,
        task: compileTask,
        map: compileMap,
        filter: compileFilter
    };
    return {
        key,
        Type: yield nodes[type](def, opts)
    };
});
exports.default = (content, instanceResolver, envs) => __awaiter(void 0, void 0, void 0, function* () {
    const definition = decode_1.default(content, envs);
    validate_1.default(schema, definition);
    let nodes = [];
    let edges = [];
    let previousKey = null;
    let i = 0;
    for (const step of definition.steps) {
        step.key = step.key || `node-${i}`;
        const opts = { defaultNodeKey: previousKey, instanceResolver };
        if (step.type === 'task' && step.inputs) {
            const mapKey = `${step.key}-inputs`;
            const mapNode = yield nodeCompiler('map', step.inputs, mapKey, opts);
            nodes.push(mapNode);
            if (previousKey) {
                edges.push({ src: previousKey, dst: mapKey });
            }
            previousKey = mapKey;
            i++;
        }
        const type = step.type !== 'trigger'
            ? step.type
            : step.eventKey
                ? 'event'
                : 'result';
        const stepNode = yield nodeCompiler(type, step, step.key, opts);
        nodes.push(stepNode);
        if (previousKey) {
            edges.push({ src: previousKey, dst: step.key });
        }
        previousKey = step.key;
        i++;
    }
    return {
        name: definition.name || definition.key,
        nodes,
        edges,
    };
});
